import type { Cookies, RequestEvent } from '@sveltejs/kit';
import { eq } from 'drizzle-orm';
import { db } from '$lib/server/db';
import { KeyCloak } from 'arctic';

import { env } from '$env/dynamic/private';
import { env as envPublic } from '$env/dynamic/public';
import { sessions, type Session } from './db/schema/sessions';
import { users } from './db/schema/users';

const DAY_IN_MS = 1000 * 60 * 60 * 24;

export const sessionCookieName = 'auth-session';

export const keycloak = new KeyCloak(
	env.KEYCLOAK_REALM || '',
	env.KEYCLOAK_CLIENT_ID || '',
	env.KEYCLOAK_CLIENT_SECRET || '',
	envPublic.PUBLIC_ENDPOINT + '/auth/callback'
);

function generateSecureRandomString(): string {
	// Human readable alphabet (a-z, 0-9 without l, o, 0, 1 to avoid confusion)
	const alphabet = 'abcdefghijklmnpqrstuvwxyz23456789';

	// Generate 24 bytes = 192 bits of entropy.
	// We're only going to use 5 bits per byte so the total entropy will be 192 * 5 / 8 = 120 bits
	const bytes = new Uint8Array(24);
	crypto.getRandomValues(bytes);

	let id = '';
	for (let i = 0; i < bytes.length; i++) {
		// >> 3 s"removes" the right-most 3 bits of the byte
		id += alphabet[bytes[i] >> 3];
	}
	return id;
}

async function hashSecret(secret: string): Promise<Uint8Array> {
	const secretBytes = new TextEncoder().encode(secret);
	const secretHashBuffer = await crypto.subtle.digest('SHA-256', secretBytes);
	return new Uint8Array(secretHashBuffer);
}

export async function createSession(user_id: string) {
	const id = generateSecureRandomString();
	const secret = generateSecureRandomString();
	const secretHash = await hashSecret(secret);

	const token = id + '.' + secret;

	const session: Session = {
		id,
		secret_hash: Buffer.from(secretHash),
		user_id,
		expires_at: new Date(Date.now() + DAY_IN_MS * 30),
		created_at: new Date(Date.now()),
		deleted_at: new Date(),
		updated_at: new Date(Date.now())
	};
	await db.insert(sessions).values(session);
	return { session, token };
}

export async function validateSessionToken(token: string) {
	const tokenParts = token.split('.');
	if (tokenParts.length !== 2) {
		return null;
	}

	const sessionId = tokenParts[0];
	const sessionSecret = tokenParts[1];

	const [result] = await db
		.select({
			// Adjust user table here to tweak returned data
			user: { id: users.id, username: users.username, email: users.email },
			session: sessions
		})
		.from(sessions)
		.innerJoin(users, eq(sessions.user_id, users.id))
		.where(eq(sessions.id, sessionId));

	if (!result) {
		return { session: null, user: null };
	}
	const { session, user } = result;

	const tokenSecretHash = await hashSecret(sessionSecret);
	const validSecret = constantTimeEqual(tokenSecretHash, new Uint8Array(session.secret_hash));
	if (!validSecret) {
		return null;
	}

	const sessionExpired = Date.now() >= session.expires_at.getTime();
	if (sessionExpired) {
		await db.delete(sessions).where(eq(sessions.id, session.id));
		return { session: null, user: null };
	}

	const renewSession = Date.now() >= session.expires_at.getTime() - DAY_IN_MS * 15;
	if (renewSession) {
		session.expires_at = new Date(Date.now() + DAY_IN_MS * 30);
		await db
			.update(sessions)
			.set({ expires_at: session.expires_at })
			.where(eq(sessions.id, session.id));
	}

	return { session, user };
}

function constantTimeEqual(a: Uint8Array, b: Uint8Array): boolean {
	if (a.byteLength !== b.byteLength) {
		return false;
	}
	let c = 0;
	for (let i = 0; i < a.byteLength; i++) {
		c |= a[i] ^ b[i];
	}
	return c === 0;
}

export type SessionValidationResult = Awaited<ReturnType<typeof validateSessionToken>>;

export async function invalidateSession(sessionId: string) {
	await db.delete(sessions).where(eq(sessions.id, sessionId));
}

export function setSessionTokenCookie(event: RequestEvent, token: string, expires_at: Date) {
	event.cookies.set(sessionCookieName, token, {
		expires: expires_at,
		path: '/'
	});
}

export function deleteSessionTokenCookie(cookies: Cookies) {
	cookies.delete(sessionCookieName, {
		path: '/'
	});
}

export function encodeSessionPublicJSON(session: Session): string {
	// Omit Session.secretHash
	const json = JSON.stringify({
		id: session.id,
		created_at: Math.floor(session.created_at.getTime() / 1000)
	});
	return json;
}
